#include <stdlib.h>
#include <string.h>

#include <sys/thread.h>

#include <sys/confnet.h>
#include <arpa/inet.h>
#include <net/if_var.h>
#include <sys/socket.h>

#include "config.h"
#include "score.h"
#include "service.h"

/*
 * Commands.
 */
static prog_char quit_cmd_P[] = "quit";
static prog_char auth_cmd_P[] = "auth";
static prog_char account_cmd_P[] = "account";
static prog_char config_cmd_P[] = "config";
static prog_char ip_cmd_P[] = "ip";
static prog_char mac_cmd_P[] = "mac";
static prog_char save_cmd_P[] = "save";
static prog_char score_cmd_P[] = "score";
static prog_char trace_cmd_P[] = "trace";

/*
 * Responses.
 */
static prog_char banner_P[] = "200 Welcome\r\n";
static prog_char eol_P[] = ".\r\n";
static prog_char ok_P[] = "200 OK\r\n";
static prog_char bad_arg_P[] = "400 Bad arg\r\n";
static prog_char bad_list_P[] = "400 Bad list\r\n";
static prog_char no_auth_P[] = "400 No authentication\r\n";
static prog_char not_found_P[] = "400 Not found\r\n";
static prog_char no_mem_P[] = "400 No mem\r\n";
static prog_char list_intro_P[] = "300 List follows\r\n";
static prog_char score_lists_P[] = "300 Score lists\r\nto\r\nfrom\r\nsubject\r\n.\r\n";
static prog_char help_P[] = "400 Commands\r\n"
              "account\t\t\t\tList accounts\r\n"
              "account <host> <user>\t\tRemove account\r\n"
              "account <host> <user> <pass>\tAdd account\r\n"
              "auth <secret>\t\t\tAuthenticate\r\n"
              "auth <old> <new>\t\tChange authentication\r\n"
              "ip\t\t\t\tList IP configuration\r\n"
              "ip addr <addr>\t\t\tSet IP address\r\n"
              "ip mask <mask>\t\t\tSet IP maks\r\n"
              "ip gate <addr>\t\t\tSet default gateway\r\n"
              "ip dns <addr>\t\t\tSet DNS\r\n"
              "mac\t\t\t\tShow MAC address\r\n"
              "mac <addr>\t\t\tSet MAC address\r\n"
              "quit\t\t\t\tExit session\r\n"
              "reset\t\t\t\tReboot system\r\n"
              "save\t\t\t\tSave configuration\r\n"
              "score\t\t\t\tList score lists\r\n"
              "score <list>\t\t\tList scores\r\n"
              "score <list> <str>\t\tRemove score\r\n"
              "score <list> <str> <val>\tAdd score\r\n"
              "score dnsbl <domain> <val>\tSet DNS block list\r\n"
              "trace <level>\t\t\tSet trace level\r\n"
              ".\r\n";


/*
 * List mail account configuration.
 */
static int ListAccounts(FILE *stream, PGM_P ann)
{
    POP3ACCOUNT *pacc = account;

    if(ann)
        fputs_P(ann, stream);
    while(pacc) {
        fprintf(stream, "account %s:%u %s\r\n", pacc->pacc_host, pacc->pacc_port, pacc->pacc_user);
        pacc = pacc->pacc_next;
    }
    if(ann)
        fputs_P(eol_P, stream);

    return 0;
}

/*
 * List IP configuration.
 *
 * No change to retrieve the DNS' IP with DHCP.
 */
static int ListIp(FILE *stream, PGM_P ann)
{
    if(ann)
        fputs_P(ann, stream);
    if(confnet.cdn_cip_addr)
        fprintf(stream, "ip addr %s\r\n", inet_ntoa(confnet.cdn_cip_addr));
    else
        fprintf(stream, "ip addr %s\r\n", inet_ntoa(confnet.cdn_ip_addr));
    fprintf(stream, "ip mask %s\r\n", inet_ntoa(confnet.cdn_ip_mask));
    fprintf(stream, "ip gate %s\r\n", inet_ntoa(confnet.cdn_gateway));
    fprintf(stream, "ip dns %s\r\n", inet_ntoa(acfg.acfg_dns));
    if(ann)
        fputs_P(eol_P, stream);

    return 0;
}

/*
 *
 */
static int ListScores(FILE *stream, PGM_P ann, char *name, SCORELIST *list)
{
    if(ann)
        fputs_P(ann, stream);
    while(list) {
        fprintf(stream, "score %s %s %d\r\n", name, list->scol_str, list->scol_val);
        list = list->scol_nxt;
    }
    if(ann)
        fputs_P(eol_P, stream);

    return 0;
}

/*
 * Split a string into arguments.
 */
static int ParseArgs(CONST char * buff, char *** argvp)
{
    char **argp;
    CONST char *cp;
    char *wp;
    int wcnt = 0;
    size_t wsiz = 0;
    u_short i;

    /*
     * Release previously allocated memory.
     */
    argp = *argvp;
    if (argp) {
	for (i = 0; argp[i]; i++)
	    free(argp[i]);
	free(argp);
        argp = 0;
    }

    /*
     * Determine number of words and largest word size.
     */
    for (cp = buff; *cp;) {
	while (*cp == ' ' || *cp == '\t')
	    cp++;
        if (*cp) {
	    i = 0;
            while (*cp && *cp != ' ' && *cp != '\t') {
	        cp++;
                i++;
            }
	    ++wcnt;
	    if (i > wsiz)
	        wsiz = i;
        }
    }


    /*
     * Create a new argument array.
     */
    if(wcnt) {
        argp = malloc((wcnt + 1) * sizeof(u_char *));
        wp = malloc(wsiz + 1);
        wcnt = 0;
        while (*buff) {
	    while (*buff == ' ' || *buff == '\t')
	        buff++;
            if (*buff) {
	        for (i = 0; *buff && *buff != ' ' && *buff != '\t';)
	            wp[i++] = *buff++;
	        wp[i] = 0;
	        argp[wcnt] = malloc(i + 1);
	        strcpy(argp[wcnt], wp);
	        wcnt++;
            }
        }
        free(wp);
        argp[wcnt] = 0;
    }
    *argvp = argp;

    return wcnt;
}



/*
 *
 */
static int ClientSession(FILE *stream)
{
    char *buff;
    char *cp;
    int auth = 0;
    char **argv = 0;
    int argc;
    PGM_P response;

    if((buff = malloc(MAX_CFGLINE)) == 0)
        return -1;
    response = banner_P;

    for(;;) {
        if(response) {
            if(fputs_P(response, stream) == EOF)
                break;
            response = 0;
        }
        fflush(stream);
        if(fgets(buff, MAX_CFGLINE, stream) == 0) 
            break;
        if((cp = strchr(buff, '\r')) != 0)
            *cp = 0;
        if((cp = strchr(buff, '\n')) != 0)
            *cp = 0;
        if(*buff == 0)
            continue;

	/* Split command buffer into arguments. */
        if((argc = ParseArgs(buff, &argv)) == 0)
	    continue;

        /*
         * quit command.
         */
        if (strcmp_P(argv[0], quit_cmd_P) == 0) {
            fputs("200 Bye\r\n", stream);
            fflush(stream);
            break;
        }

	/*
         * auth command.
         */
	if (strcmp_P(argv[0], auth_cmd_P) == 0) {
            if(argc >= 2)
                auth = strcmp(acfg.acfg_secret, argv[1]) == 0;
            if(auth) {
                if(argc == 3) {
                    if(strlen(argv[2]) >= sizeof(acfg.acfg_secret)) {
                        response = bad_arg_P;
                        continue;
                    }
                    strcpy(acfg.acfg_secret, argv[2]);
                }
                response = ok_P;
                continue;
            }
        }

        /*
         * All commands below this point require authentication.
         */
        if(auth == 0) {
            response = no_auth_P;
            continue;
        }

	/*
         * account command.
         */
	if (strcmp_P(argv[0], account_cmd_P) == 0) {
            if(argc == 1)
                ListAccounts(stream, list_intro_P);
            else if(argc == 3) {
                if(DelAccount(argv[1], argv[2]))
                    response = not_found_P;
                else
                    response = ok_P;
            }
            else if(argc == 4) {
                if(AddAccount(argv[1], argv[2], argv[3], POP3PORT, 0))
                    response = no_mem_P;
                else
                    response = ok_P;
            }
            else
                response = bad_arg_P;
            continue;
        }

	/*
         * config command.
         */
	if (strcmp_P(argv[0], config_cmd_P) == 0) {
            u_char *mcp = confnet.cdn_mac;

            fputs_P(list_intro_P, stream);
            ListAccounts(stream, 0);
            ListIp(stream, 0);
            fprintf(stream, "mac %02X%02X%02X%02X%02X%02X\r\n", 
                        mcp[0], mcp[1], mcp[2], mcp[3], mcp[4], mcp[5]);
            if(dnsbl_score)
                fprintf(stream, "score dnsbl %s %d\r\n", dnsbl_domain, dnsbl_score);
            ListScores(stream, 0, "to", list_to);
            ListScores(stream, 0, "from", list_from);
            ListScores(stream, 0, "subject", list_subject);
            response = eol_P;
            continue;
        }

	/*
         * ip command.
         */
	if (strcmp_P(argv[0], ip_cmd_P) == 0) {
            if(argc == 1) 
                ListIp(stream, list_intro_P);
            else if(argc == 3) {
                if(SetIp(argv[1], argv[2]))
                    response = bad_arg_P;
                else
                    response = ok_P;
            }
            else
                response = bad_arg_P;
            continue;
        }

	/*
         * mac command.
         */
	if (strcmp_P(argv[0], mac_cmd_P) == 0) {
            if(argc == 1) {
                u_char *mcp = confnet.cdn_mac;
                fprintf(stream, "200 mac %02X%02X%02X%02X%02X%02X\r\n", 
                        mcp[0], mcp[1], mcp[2], mcp[3], mcp[4], mcp[5]);
                continue;
            }
            else if(argc == 2) {
                if(SetMac(argv[1]))
                    response = bad_arg_P;
                else
                    response = ok_P;
            }
            else
                response = bad_arg_P;
            continue;
        }

	/*
         * save command.
         */
	if (strcmp_P(argv[0], save_cmd_P) == 0) {
            ConfigSave();
            response = ok_P;
            continue;
        }

	/*
         * score command.
         */
	if (strcmp_P(argv[0], score_cmd_P) == 0) {
            if(argc >= 2) {
                SCORELIST **lpp;

                if(strcmp(argv[1], "dnsbl") == 0) {
                    if(argc == 2)
                        fprintf(stream, "score dnsbl %s %d\r\n", dnsbl_domain, dnsbl_score);
                    else if(argc >= 3 && strlen(argv[2]) < sizeof(dnsbl_domain)) {
                        strcpy(dnsbl_domain, argv[2]);
                        if(argc == 4) {
                            if((dnsbl_score = atoi(argv[3])) > 0)
                                dnsbl_score = -dnsbl_score;
                        }
                        else
                            dnsbl_score = 0;
                        fputs_P(ok_P, stream);
                    }
                    else
                        fputs_P(bad_arg_P, stream);
                    continue;
                }
                if(argv[1][0] == 't')
                    lpp = &list_to;                    
                else if(argv[1][0] == 'f')
                    lpp = &list_from;
                else if(argv[1][0] == 's')
                    lpp = &list_subject;
                else {
                    response = bad_list_P;
                    continue;
                }

                if(argc == 2) 
                    ListScores(stream, list_intro_P, argv[1], *lpp);
                else if(argc == 3) {
                    if(DelScore(argv[2], lpp))
                        response = not_found_P;
                    else
                        response = ok_P;
                }
                else if(argc == 4) {
                    if(AddScore(argv[2], atoi(argv[3]), lpp))
                        response = no_mem_P;
                    else
                        response = ok_P;
                }
                else
                    response = bad_arg_P;
            }
            else
                response = score_lists_P;
            continue;
        }

	/*
         * trace command.
         */
	if (strcmp_P(argv[0], trace_cmd_P) == 0) {
            if(argc == 2) {
                trace_level = atoi(argv[1]);
                response = ok_P;
            }
            else
                response = bad_arg_P;
            continue;
        }

        /*
         * Anything else invokes help.
         */
        response = help_P;
    }

    /*
     * Release allocated memory.
     */
    ParseArgs("", &argv);
    free(buff);

    return 0;
}


THREAD(SerialTalk, arg)
{
    FILE *stream = arg;

    for(;;)
        ClientSession(stream);
}

THREAD(TelnetTalk, arg)
{
    TCPSOCKET *sock;
    FILE *stream;
    u_long tmo = 30000;

    /*
     * Now loop endless for connections.
     */
    for(;;) {
        /*
         * Create a socket.
         */
	if ((sock = NutTcpCreateSocket()) != 0) {
	    /*
	     * Listen on port 23. If we return, we got a client.
	     */
	    if (NutTcpAccept(sock, 23) == 0) {
                NutTcpSetSockOpt(sock, SO_RCVTIMEO, &tmo, sizeof(tmo));

	        /*
		 * Open a stream and associate it with the socket, so 
		 * we can use standard I/O. Note, that socket streams
		 * currently do support text mode.
	         */
	        if((stream = _fdopen((int)sock, "r+b")) != 0) {
	            /*
	             * Process client requests.
	             */
	    	    ClientSession(stream);

		    /*
		     * Close the stream.
		     */
		    fclose(stream);
		}
	    }

            /*
	     * Close our socket.
	     */
	    NutTcpCloseSocket(sock);
	}
    }
}

int ServiceStart(FILE *stream, int network)
{
    if(network)
        NutThreadCreate("ttalk", TelnetTalk, 0, 512);
    if(stream)
        NutThreadCreate("stalk", SerialTalk, stream, 512);

    return 0;
}
